//======================================================================
//
// CacheMan device driver
//
// Copyright (C) 1997 Bryce Cogswell and Mark Russinovich
// http://www.ntinternals.com
//
// This program requires an NTExport generated library of unexported
// cache manager functions and variables.
//
//======================================================================
#include <ntddk.h>
#include <stdarg.h>
#include <stdio.h>
#include "ioctlcmd.h"
#include "cacheman.h"

//======================================================================
//                           D E F I N E S
//======================================================================
// print macro that only turns on when debugging is on
#if DBG
#define DbgPrint(arg) DbgPrint arg
#else
#define DbgPrint(arg) 
#endif


//======================================================================
//              E X T E R N S   A N D   F O R W A R D S  
//======================================================================

//
// NTExport library initialization routine
// 
NTSTATUS		ExportInit( void );

//
// Extern definitions for functions and variables exported by the
// NTExport generated library
// 
extern PULONG		CcDirtyPageThreshold;
extern PULONG		CcDirtyPageTarget;
extern PULONG		CcMaxDirtyWrite;
extern PULONG		CcAvailablePagesThreshold;
extern PLARGE_INTEGER	CcFirstDelay;
extern PLARGE_INTEGER	CcIdleDelay;
extern PLARGE_INTEGER	CcCollisionDelay;
extern PLARGE_INTEGER	CcTargetCleanDelay;
extern struct {
    UCHAR		dummy[ 0x14 ];
    ULONG		MmSystemCacheWsSize;	/* fabricated name */
    ULONG		MmSystemCacheWsMinimum;	/* shadows */
    ULONG		MmSystemCacheWsMaximum;	/* shadows */
}		*	MmSystemCacheWs;
extern PULONG		MmWorkingSetReductionMaxCacheWs;
extern PULONG		MmWorkingSetReductionMinCacheWs;
extern PULONG		MmWorkingSetVolReductionMaxCacheWs;
extern PUCHAR		MiDoPeriodicAgressiveTrimming;
extern PULONG		MmPeriodicAgressiveTrimMinFree;
extern PULONG		MmPeriodicAgressiveTrimMaxFree;
extern PULONG		MmPeriodicAgressiveCacheWsMin;


//======================================================================
//                           G L O B A L S 
//======================================================================

//
// Our user-inteface device object
//
PDEVICE_OBJECT          GUIDevice;

//
// Have we initialized the export library?
//
BOOLEAN                 LibraryInitialized = FALSE;


//======================================================================
//         D E V I C E - D R I V E R  R O U T I N E S 
//======================================================================


//----------------------------------------------------------------------
//
// DeviceControl
//
//----------------------------------------------------------------------
BOOLEAN DeviceControl( IN PFILE_OBJECT FileObject,
		       IN BOOLEAN Wait,
		       IN PVOID InputBuffer, IN ULONG InputBufferLength, 
		       OUT PVOID OutputBuffer, IN ULONG OutputBufferLength, 
		       IN ULONG IoControlCode,
		       OUT PIO_STATUS_BLOCK IoStatus, 
		       IN PDEVICE_OBJECT DeviceObject )
{
    NTSTATUS			status = STATUS_SUCCESS;
    struct cacheman_getset    *	getset;

    //
    // Process the device I/O control, assuming success.
    //
    IoStatus->Status      = STATUS_SUCCESS;
    IoStatus->Information = 0;  

    switch ( IoControlCode ) {

	case CACHEMAN_GET:

            //
            // Gui is requesting current cache settings.
            // Make sure buffer is large enough.
            //
	    if ( OutputBufferLength != sizeof *getset ) {

		status = STATUS_INVALID_PARAMETER;
		break;
	    }

            //
            // Initialize export library 
            //
            if( !LibraryInitialized ) {

                status = ExportInit();
                if ( !NT_SUCCESS(status) ) {

                    break;
                }
            
                LibraryInitialized = TRUE;
            }

            //
            // Point at output buffer and fill in fields
            //
	    getset = OutputBuffer;

	    getset->CcDirtyPageThreshold	= *CcDirtyPageThreshold;
	    getset->CcDirtyPageTarget		= *CcDirtyPageTarget;
	    getset->CcMaxDirtyWrite		= *CcMaxDirtyWrite;
	    getset->CcAvailablePagesThreshold	= *CcAvailablePagesThreshold;
	    getset->CcFirstDelay		= *CcFirstDelay;
	    getset->CcIdleDelay			= *CcIdleDelay;
	    getset->CcCollisionDelay		= *CcCollisionDelay;
	    getset->CcTargetCleanDelay		= *CcTargetCleanDelay;

	    getset->MmSystemCacheWsMinimum
	          = MmSystemCacheWs->MmSystemCacheWsMinimum;
	    getset->MmSystemCacheWsMaximum
	          = MmSystemCacheWs->MmSystemCacheWsMaximum;

	    getset->MmWorkingSetReductionMaxCacheWs
	         = *MmWorkingSetReductionMaxCacheWs;
	    getset->MmWorkingSetReductionMinCacheWs
	         = *MmWorkingSetReductionMinCacheWs;
	    getset->MmWorkingSetVolReductionMaxCacheWs
	         = *MmWorkingSetVolReductionMaxCacheWs;

	    getset->MiDoPeriodicAgressiveTrimming
	         = *MiDoPeriodicAgressiveTrimming;
	    getset->MmPeriodicAgressiveTrimMinFree
	         = *MmPeriodicAgressiveTrimMinFree;
	    getset->MmPeriodicAgressiveTrimMaxFree
	         = *MmPeriodicAgressiveTrimMaxFree;
	    getset->MmPeriodicAgressiveCacheWsMin
	         = *MmPeriodicAgressiveCacheWsMin;

	    IoStatus->Information = sizeof *getset;
	    break;


	case CACHEMAN_SET:

            //            
            // Gui is sending down new cache settings
            //            
	    if ( InputBufferLength != sizeof *getset ) {
		status = STATUS_INVALID_PARAMETER;
		break;
	    }
	    getset = InputBuffer;

	    *CcDirtyPageThreshold	= getset->CcDirtyPageThreshold;
	    *CcDirtyPageTarget		= getset->CcDirtyPageTarget;
	    *CcMaxDirtyWrite		= getset->CcMaxDirtyWrite;
	    *CcAvailablePagesThreshold	= getset->CcAvailablePagesThreshold;
	    *CcFirstDelay		= getset->CcFirstDelay;
	    *CcIdleDelay		= getset->CcIdleDelay;
	    *CcCollisionDelay		= getset->CcCollisionDelay;
	    *CcTargetCleanDelay		= getset->CcTargetCleanDelay;

	    MmSystemCacheWs->MmSystemCacheWsMinimum
	      = getset->MmSystemCacheWsMinimum;
	    MmSystemCacheWs->MmSystemCacheWsMaximum
	      = getset->MmSystemCacheWsMaximum;

	    *MmWorkingSetReductionMaxCacheWs
	      = getset->MmWorkingSetReductionMaxCacheWs;
	    *MmWorkingSetReductionMinCacheWs
	      = getset->MmWorkingSetReductionMinCacheWs;
	    *MmWorkingSetVolReductionMaxCacheWs
	      = getset->MmWorkingSetVolReductionMaxCacheWs;
	    
	    *MiDoPeriodicAgressiveTrimming
	      = getset->MiDoPeriodicAgressiveTrimming != 0;
	    *MmPeriodicAgressiveTrimMinFree
	      = getset->MmPeriodicAgressiveTrimMinFree;
	    *MmPeriodicAgressiveTrimMaxFree
	      = getset->MmPeriodicAgressiveTrimMaxFree;
	    *MmPeriodicAgressiveCacheWsMin
	      = getset->MmPeriodicAgressiveCacheWsMin;

	    IoStatus->Information = 0;
	    break;

        default:
	    status = STATUS_INVALID_PARAMETER;
	    break;
    }

    IoStatus->Status = status;
    return NT_SUCCESS( status );
}


//----------------------------------------------------------------------
//
// Dispatch
//
// In this routine we handle requests to our own device. The only 
// requests we care about handling explicitely are IOCTL commands that
// we will get from the GUI. We also expect to get Create and Close 
// commands when the GUI opens and closes communications with us.
//
//----------------------------------------------------------------------
NTSTATUS Dispatch( IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp )
{
    PIO_STACK_LOCATION      irpStack;
    PVOID                   inBuffer;
    PVOID                   outBuffer;
    ULONG                   inBufferLength;
    ULONG                   outBufferLength;
    ULONG                   ioControlCode;
    
    //
    // Go ahead and set the request up as successful 
    //
    Irp->IoStatus.Status      = STATUS_SUCCESS;
    Irp->IoStatus.Information = 0;

    //
    // Get a pointer to the current location in the Irp. This is where 
    // the function codes and parameters are located.  
    // 
    irpStack = IoGetCurrentIrpStackLocation (Irp);

    //
    // Get the pointer to the input/output buffer and its length 
    //
    inBuffer        = Irp->AssociatedIrp.SystemBuffer;
    inBufferLength  = irpStack->Parameters.DeviceIoControl.InputBufferLength;
    outBuffer       = Irp->AssociatedIrp.SystemBuffer;;
    outBufferLength = irpStack->Parameters.DeviceIoControl.OutputBufferLength;
    ioControlCode   = irpStack->Parameters.DeviceIoControl.IoControlCode;

    switch (irpStack->MajorFunction) {
	case IRP_MJ_DEVICE_CONTROL:
		DbgPrint (("CacheMan: IRP_MJ_DEVICE_CONTROL\n"));
		/* Its a request from the GUI */
		DeviceControl( irpStack->FileObject, TRUE,
				  inBuffer, inBufferLength, 
				  outBuffer, outBufferLength,
				  ioControlCode, &Irp->IoStatus,
				  DeviceObject );
		break;
    }
    IoCompleteRequest( Irp, IO_NO_INCREMENT );
    return STATUS_SUCCESS;   
}


//----------------------------------------------------------------------
//
// Unload
//
// Our job is done - time to leave.
//
//----------------------------------------------------------------------
VOID Unload( IN PDRIVER_OBJECT DriverObject )
{
    WCHAR		deviceLinkBuffer[]  = L"\\DosDevices\\cacheman";
    UNICODE_STRING	deviceLinkUnicodeString;
    PDEVICE_OBJECT	DeviceObject;

    DbgPrint(("cacheman.SYS: unloading\n"));

    //
    // Delete the symbolic link for our device 
    //
    RtlInitUnicodeString( &deviceLinkUnicodeString, deviceLinkBuffer );
    IoDeleteSymbolicLink( &deviceLinkUnicodeString );

    //
    // Delete the device objects 
    //
    DeviceObject = DriverObject->DeviceObject;

    while ( DeviceObject != NULL ) {

	PDEVICE_OBJECT	Prev = DeviceObject;
	DeviceObject = DeviceObject->NextDevice;
	IoDeleteDevice( Prev );
    }

    DbgPrint(("cacheman.SYS: deleted devices\n"));
}


//----------------------------------------------------------------------
//
// DriverEntry
//
// Installable driver initialization. Here we just set ourselves up.
//
//----------------------------------------------------------------------
NTSTATUS DriverEntry( IN PDRIVER_OBJECT DriverObject,
		      IN PUNICODE_STRING RegistryPath )
{
    NTSTATUS                ntStatus;
    WCHAR                   deviceNameBuffer[]  = L"\\Device\\cacheman";
    UNICODE_STRING          deviceNameUnicodeString;
    WCHAR                   deviceLinkBuffer[]  = L"\\DosDevices\\cacheman";
    UNICODE_STRING          deviceLinkUnicodeString;

    DbgPrint (("cacheman.SYS: entering DriverEntry\n"));

    //
    // Setup our name 
    //
    RtlInitUnicodeString( &deviceNameUnicodeString, deviceNameBuffer );

    //
    // Set up the device used for GUI communications 
    //
    ntStatus = IoCreateDevice ( DriverObject,
				0,
				&deviceNameUnicodeString,
				FILE_DEVICE_CACHEMAN,
				0,
				TRUE,
				&GUIDevice );
    if ( NT_SUCCESS(ntStatus) ) {

	//
        // Create a symbolic link that the GUI can specify to gain access 
	// to this driver/device    
        //
	RtlInitUnicodeString( &deviceLinkUnicodeString,
			      deviceLinkBuffer );
	ntStatus = IoCreateSymbolicLink( &deviceLinkUnicodeString,
					 &deviceNameUnicodeString );
	if ( ! NT_SUCCESS(ntStatus) ) {

	    DbgPrint (("cacheman.SYS: IoCreateSymbolicLink failed\n"));
        }

	//
        // Create dispatch points for all routines that must be handled 
        //
	DriverObject->MajorFunction[IRP_MJ_CREATE]          =
	DriverObject->MajorFunction[IRP_MJ_CLOSE]           =
	DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL]  = Dispatch;
	DriverObject->DriverUnload                          = Unload;

    }

    //
    // If any error at all, cleanup on the way out
    //
    if ( ! NT_SUCCESS(ntStatus) ) {

	DbgPrint(("CacheMan: Failed to create our device!\n"));
	if ( GUIDevice ) {

	    IoDeleteDevice( GUIDevice );
        }
	return ntStatus;
    }

    return ntStatus;
}
    
